/*
	Integrate.c
		integrate over the unit hypercube
		this file is part of Vegas
		last modified 17 Dec 07 th
*/


// do nothing...
//---------------------
// signal catching
#include "catchSignal.h"
#include "catchSignal.c"
//---------------------



static int Integrate(creal epsrel, creal epsabs,
		     //  cint flags, cnumber mineval, cnumber maxeval,
		     cint flags, cnumber thermsteps, cnumber maxeval,
		     cnumber nstart, cnumber nincrease,cnumber ncadapt,
		     real *integral, real *error, real *prob)
{

#ifdef CATCHSIG
  bool terminate = false;
  csignal=0;
  // invoke signals
  (void) signal(SIGINT, ex_program);
  (void) signal(SIGTERM, ex_program);
  (void) signal(SIGXCPU, ex_program);
  (void) signal(SIGUSR2, ex_program);
  // other signals left as default

#endif


  cnumber mineval=10;

  real *sample;
  count dim, comp;
  int fail = 1;
  struct {
    count niter;
    number nsamples, neval;
    Cumulants cumul[NCOMP];
    Grid grid[NDIM];
  } state;
  int statemsg = VERBOSE;
  struct stat st;




  if( VERBOSE > 1 ) {
    char s[256];
    sprintf(s, "Vegas input parameters:\n"
      "  ndim " COUNT "\n  ncomp " COUNT "\n"
      "  epsrel " REAL "\n  epsabs " REAL "\n"
      "  flags %d\n  thermsteps " NUMBER "\n  maxeval " NUMBER "\n"
      "  nstart " NUMBER "\n  nincrease " NUMBER "\n"
      "  vegasgridno %d\n  vegasstate \"%s\"\n",
      ndim_, ncomp_,
      epsrel, epsabs,
      flags, thermsteps, maxeval,
      nstart, nincrease,
      EXPORT(vegasgridno), EXPORT(vegasstate));
    Print(s);
  }

#ifdef MLVERSION
  if( setjmp(abort_) ) goto abort;
#endif

  char s[256];
  sprintf(s,"Initializing RNG... %d",sizeof(number)*8);
  Print(s);
  IniRandom(2*maxeval, flags);
  Print("RNG initialized...");

  if( *EXPORT(vegasstate) && stat(EXPORT(vegasstate), &st) == 0 &&
      st.st_size == sizeof(state) && (st.st_mode & 0400) ) {
    cint h = open(EXPORT(vegasstate), O_RDONLY);
    read(h, &state, sizeof(state));
    close(h);
    SkipRandom(neval_ = state.neval);

    if( VERBOSE ) {
      char s[256];
      sprintf(s, "\n [INFO] Restoring state from %s.", EXPORT(vegasstate));
      Print(s);
    }
  }
  else {
    state.niter = 0;
    state.nsamples = nstart;
    Zap(state.cumul);
    GetGrid(state.grid);
  }

  SamplesAlloc(sample, EXPORT(vegasnbatch));

  /* main iteration loop */

  for( ; ; ) {

    // clear results after thermalization iterations
    if(state.niter==thermsteps) {
      char s[256];
      sprintf(s,"\n [INFO] Thermalization Iterations (%d) done. Clearing Cumulants.\n",state.niter);
      Print(s);
      for( comp = 0; comp < ncomp_; ++comp )
	flushCumulant(&state.cumul[comp]);
    }      

#ifdef CATCHSIG
    if(csignal==SIGINT || terminate) break;
#endif

    number nsamples = state.nsamples;
    creal jacobian = 1./nsamples;
    Grid margsum[NCOMP][NDIM];

    Zap(margsum);

    for( ; nsamples > 0; nsamples -= EXPORT(vegasnbatch) ) {
      cnumber nbatch = IMin(EXPORT(vegasnbatch), nsamples);
      real *w = sample;
      real *x = w + nbatch;
      real *f = x + nbatch*ndim_;
      real *lastf = f + nbatch*ncomp_;
      bin_t *bin = (bin_t *)lastf;

      while( x < f ) {
        real weight = jacobian;

        GetRandom(x);

        for( dim = 0; dim < ndim_; ++dim ) {
          creal pos = *x*NBINS;
          ccount ipos = (count)pos;
          creal prev = (ipos == 0) ? 0 : state.grid[dim][ipos - 1];
          creal diff = state.grid[dim][ipos] - prev; 
          *x++ = prev + (pos - ipos)*diff;
          *bin++ = ipos;
          weight *= diff*NBINS;
        }

        *w++ = weight;
      }

      DoSample(nbatch, sample, w, f);
      
#ifdef CATCHSIG    
      if(csignal == SIGTERM || csignal == SIGUSR2 || csignal == SIGXCPU) {
	terminate=true;
	break;
      }
#endif

      w = sample;
      bin = (bin_t *)lastf;

      while( f < lastf ) {
        creal weight = *w++;

        for( comp = 0; comp < ncomp_; ++comp ) {
          real wfun = weight*(*f++);
          if( wfun ) {
            Cumulants *c = &state.cumul[comp];
            Grid *m = margsum[comp];

            c->sum += wfun;
            c->sqsum += wfun *= wfun;
            for( dim = 0; dim < ndim_; ++dim )
              m[dim][bin[dim]] += wfun;
          }
        }

        bin += ndim_;
      }
    }

    // only if signal is fatal (TERM, USR2, XCPU), BREAK from loop
#ifdef CATCHSIG    
    if(terminate) break;
#endif
      
    fail = 0;

    /* compute the integral and error values */

    double thisIterRes = 0.0;
    double thisIterErr = 0.0;

    for( comp = 0; comp < ncomp_; ++comp ) {
      Cumulants *c = &state.cumul[comp];

      real avg, sigsq;
      real w = Weight(c->sum, c->sqsum, state.nsamples);

      if(comp==0) {
	thisIterRes=c->sum; 
	thisIterErr=sqrt(1./w);
      }

      sigsq = 1/(c->weightsum += w);
      avg = sigsq*(c->avgsum += w*c->sum);


      c->avg = LAST ? (sigsq = 1/w, c->sum) : avg;
      c->err = sqrt(sigsq);
      
      // check termination criterium only for main integrand
      if(comp==0) {
	fail |= (c->err > MaxErr(c->avg));
        // if we are still in the thermalization steps, we never want to quit
	if(state.niter<=thermsteps)
	  fail=1;
      }

      if( state.niter == 0 ) c->guess = c->sum;
      else {
        c->chisum += w *= c->sum - c->guess;
        c->chisqsum += w*c->sum;
      }
      c->chisq = c->chisqsum - avg*c->chisum;

      c->sum = c->sqsum = 0;
    }

    if( VERBOSE ) {

      bool useNewVerbosity = true;

      if( useNewVerbosity ) {
	char s[768], *p = s;
	cCumulants *c = &state.cumul[0];
	p += sprintf(p,"\n Iteration #%4d :\n",state.niter+1);
	p += sprintf(p,"%-20s %c %-14.7g +/- %-9.2g\n","     sigma",'=',thisIterRes,thisIterErr);
	p += sprintf(p,"%-20s %c %-14.7g +/- %-9.2g\n","     <sigma>",'=',c->avg,c->err);
	p += sprintf(p,"%-20s %c %-4.2g %c\n","     rel. error",'=',100.0*c->err/c->avg,'%');
	p += sprintf(p,"%-20s %c %-9.2g\n","     chi^2",'=',c->chisq);
	p += sprintf(p,"%-20s %c %-8d\n","     # evaluations",'=', neval_);
	
	Print(s);
	
      } else {
	char s[128 + 128*NCOMP], *p = s;	
	p += sprintf(p, "\n"
		     "Iteration " COUNT ":  " NUMBER " integrand evaluations so far",
		     state.niter + 1, neval_);
	
	for( comp = 0; comp < ncomp_; ++comp ) {
	  cCumulants *c = &state.cumul[comp];
	  p += sprintf(p, "\n[" COUNT "] "
		       REAL " +- " REAL "  \tchisq " REAL " (" COUNT " df)",
		       comp + 1, c->avg, c->err, c->chisq, state.niter);
	}
	
	Print(s);
      }
    }

    /* move ths down, such that the grid gets written also if the term. crits. are fulfilled 
    if( fail == 0 && neval_ >= mineval ) {
      // do *NOT* delete the grid-file
      //if( *EXPORT(vegasstate) ) unlink(EXPORT(vegasstate));
      break;
    }

    if( neval_ >= maxeval && *EXPORT(vegasstate) == 0 ) break;
    */


    if( ncomp_ == 1  )
      for( dim = 0; dim < ndim_; ++dim )
        RefineGrid(state.grid[dim], margsum[0][dim], flags);
    else {
      for( dim = 0; dim < ndim_; ++dim ) {
        Grid wmargsum;
        Zap(wmargsum);
	int nc;
	if( ADAPTTOFIRST  )
	  nc=1;
	else {
	  //nc=ncomp_;
	  nc=ncadapt;
	}
        for( comp = 0; comp < nc; ++comp ) {
          real w = state.cumul[comp].avg;
          if( w != 0 ) {
            creal *m = margsum[comp][dim];
            count bin;
            w = 1/Sq(w);
            for( bin = 0; bin < NBINS; ++bin )
              wmargsum[bin] += w*m[bin];
          }
        }
        RefineGrid(state.grid[dim], wmargsum, flags);
      }
    }

    ++state.niter;

    state.nsamples += nincrease;

    if( *EXPORT(vegasstate) ) {
      cint h = creat(EXPORT(vegasstate), 0666);
      if( h != -1 ) {
        state.neval = neval_;
        write(h, &state, sizeof(state));
        close(h);

        if( statemsg ) {
          char s[256];
          sprintf(s, "\n [INFO] Saving state to %s.", EXPORT(vegasstate));
          Print(s);
          statemsg = false;
        }
      }
      if( neval_ >= maxeval && maxeval > 0) {
	char s[256];
	sprintf(s,"\n [INFO] Reached Max. Evaluations of %-8d\n",neval_);
	Print(s);
	break;}
    }

    if( fail == 0 && neval_ >= mineval ) {
      char s[256];
      sprintf(s,"\n [INFO] Reached Accuracy Limit for main integrand.\n");
      Print(s);
      break;}

  } // END OF MAIN LOOP, WHATEVER COMES AFTER, WE WANT TO DO ANYWAYS

  for( comp = 0; comp < ncomp_; ++comp ) {
    cCumulants *c = &state.cumul[comp];
    integral[comp] = c->avg;
    error[comp] = c->err;

    if(state.niter<=thermsteps)
      prob[comp] = ChiSquare(c->chisq, state.niter);
    else
      prob[comp] = ChiSquare(c->chisq, state.niter-thermsteps);    
  }

  char snew[256];
  sprintf(snew,"\n [INFO] Probability that the Error estimation is wrong P = %-3.2f %c.\n",100.*prob[0],'%');
  Print(snew);

#ifdef MLVERSION
abort:
#endif

  free(sample);
  PutGrid(state.grid);

  return fail;
}

